#include <SDL.h>
#include <assert.h>
#include <algorithm>
#include <iostream>
#include <stdexcept>
#include "button_input.hpp"

namespace oiran { namespace linda {


	button_input_t::button_input_t()
	{
		std::fill(M_buttons, M_buttons + number_of_buttons, 0);
	}

	button_input_t::~button_input_t()
	{
	}

	void
	button_input_t::_update_buttons(unsigned int state)
	{
		for(int i = 0; i < number_of_buttons; ++i)
		{
			if(state & (1 << i))
				M_buttons[i] = std::max(1, M_buttons[i] + 1);
			else
				M_buttons[i] = std::min(0, M_buttons[i] - 1);
		}
	}


	const int joystick_t::axis_play = 1 << 14;


	joystick_t::joystick_t() :
			M_stick(0),
			M_out(0)
	{
		M_replay_data.state = 0;
		M_replay_data.repeat = 0;
	}

	joystick_t::~joystick_t()
	{
		stop_record();
		if(M_stick)
		{
			SDL_JoystickClose(M_stick);
#ifndef NDEBUG
			std::cout << "close joystick." << std::endl;
#endif
		}
	}

	void
	joystick_t::init()
	{
		assert(M_stick == 0 && "joystick is already opened.");

		if(!SDL_WasInit(SDL_INIT_JOYSTICK))
			if(SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0)
				throw std::runtime_error(SDL_GetError());
		M_stick = SDL_JoystickOpen(0);
#ifndef NDEBUG
		if(M_stick)
			std::cout << "open joystick." << std::endl;
#endif
	}

	void
	joystick_t::update()
	{
		unsigned int state = 0;

		if(M_stick)
		{
			SDL_JoystickUpdate();
			int x = SDL_JoystickGetAxis(M_stick, 0);
			int y = SDL_JoystickGetAxis(M_stick, 1);
			if(y < -axis_play) state |= 1 << up;
			if(y > axis_play) state |= 1 << down;
			if(x < -axis_play) state |= 1 << left;
			if(x > axis_play) state |= 1 << right;
			for(int i = a_button; i < number_of_buttons; ++i)
				if(SDL_JoystickGetButton(M_stick, i - a_button))
					state |= 1 << i;
		}

		Uint8* keys = SDL_GetKeyState(0);
		if(keys[SDLK_UP] || keys[SDLK_KP8] || keys[SDLK_i] || keys[SDLK_w])
			state |= 1 << up;
		if(keys[SDLK_DOWN] || keys[SDLK_KP2] || keys[SDLK_k] || keys[SDLK_s])
			state |= 1 << down;
		if(keys[SDLK_LEFT] || keys[SDLK_KP4] || keys[SDLK_j] || keys[SDLK_a])
			state |= 1 << left;
		if(keys[SDLK_RIGHT] || keys[SDLK_KP6] || keys[SDLK_l] || keys[SDLK_d])
			state |= 1 << right;
		if(keys[SDLK_z] || keys[SDLK_COMMA] ||
				keys[SDLK_RCTRL] || keys[SDLK_LCTRL])
			state |= 1 << a_button;
		if(keys[SDLK_x] || keys[SDLK_PERIOD] ||
				keys[SDLK_RSHIFT] || keys[SDLK_LSHIFT])
			state |= 1 << b_button;
		if(keys[SDLK_c] || keys[SDLK_SLASH] ||
				keys[SDLK_RALT] || keys[SDLK_LALT] ||
				keys[SDLK_RMETA] || keys[SDLK_LMETA])
			state |= 1 << c_button;

		_update_buttons(state);

		if(M_out)
		{
			if(M_replay_data.state != state && M_replay_data.repeat)
			{
				M_out->write(reinterpret_cast<const char*>(&M_replay_data),
						sizeof(M_replay_data));
				if(M_out->fail())
				{
					M_out->clear();
					M_out = 0;
					throw std::runtime_error("failed to save replay.");
				}
				M_replay_data.state = state;
				M_replay_data.repeat = 0;
			}
			++M_replay_data.repeat;
		}
	}

	void
	joystick_t::start_record(std::ostream& out)
	{
		M_out = &out;
		M_replay_data.state = 0;
		M_replay_data.repeat = 0;
	}

	void
	joystick_t::stop_record()
	{
		if(M_out)
		{
			M_out->write(reinterpret_cast<const char*>(&M_replay_data),
					sizeof(M_replay_data));
			if(M_out->fail())
			{
				M_out->clear();
				M_out = 0;
				throw std::runtime_error("failed to save replay.");
			}
			M_out = 0;
		}
	}


	button_replay_t::button_replay_t() :
			M_in(0)
	{
		M_replay_data.state = 0;
		M_replay_data.repeat = 0;
	}

	void
	button_replay_t::start_replay(std::istream& in)
	{
		M_in = &in;
		M_in->read(reinterpret_cast<char*>(&M_replay_data),
				sizeof(M_replay_data));
		if(M_in->fail())
		{
			M_in->clear();
			M_in = 0;
			M_replay_data.state = 0;
			M_replay_data.repeat = 0;
		}
	}

	void
	button_replay_t::stop_replay()
	{
		M_in = 0;
		M_replay_data.state = 0;
		M_replay_data.repeat = 0;
	}

	void
	button_replay_t::update()
	{
		unsigned int state = M_replay_data.state;

		if(M_in && M_replay_data.repeat > 0 && --M_replay_data.repeat == 0)
		{
			M_in->read(reinterpret_cast<char*>(&M_replay_data),
					sizeof(M_replay_data));
			if(M_in->fail())
			{
				M_in->clear();
				M_in = 0;
				M_replay_data.state = 0;
				M_replay_data.repeat = 0;
			}
			state = M_replay_data.state;
		}

		_update_buttons(state);
	}


}} // namespace oiran::linda.
